const std = @import("std");
const fs = std.fs;
const io = std.io;
const os = std.os;
const mem = std.mem;
const Options = @import("options.zig").Options;
const program_name = @import("options.zig").program_name;
const help_text = @import("options.zig").help_text;
const toUpper = @import("to_upper.zig").toUpper;
const ps = fs.path.sep_str;
pub const dict_file = "en.txt";

const default_dictionary =
    ".." ++ ps ++ "share" ++ ps ++ program_name ++ ps ++ dict_file;
var mem_buf: [0x10000]u8 = undefined;

pub fn main() !void {
    var fixed_alloc = std.heap.FixedBufferAllocator.init(&mem_buf);
    var allocator = fixed_alloc.allocator();
    const args = try std.process.argsAlloc(allocator);
    defer std.process.argsFree(allocator, args);
    const std_out = io.getStdOut();
    const writer = std_out.writer();
    var buffered = io.bufferedWriter(writer);
    var buff_writer = buffered.writer();

    var options = Options{
        .add_newline = std_out.isTty(),
    };

    options.parse(args[1..]) catch |err| {
        return switch (err) {
            error.Help => try writer.writeAll(help_text),
            else => err,
        };
    };

    var seed_bytes: [std.rand.DefaultCsprng.secret_seed_length]u8 = undefined;
    std.crypto.random.bytes(&seed_bytes);
    var gimli = std.rand.DefaultCsprng.init(seed_bytes);
    var prng = gimli.random();

    if (options.generate_password) {
        try writePassword(&buff_writer, &prng, options.letters);
    } else {
        const map = try mapFile(allocator, options);
        defer os.munmap(map);
        try writePassphrase(&buff_writer, &prng, map, options);
    }

    if (options.add_newline)
        try buff_writer.writeByte('\n');

    try buffered.flush();
}

fn mapFile(allocator: mem.Allocator, options: Options) ![]align(mem.page_size) u8 {
    var buf: [fs.MAX_PATH_BYTES]u8 = undefined;
    const file_path = options.dictionary orelse try fs.path.join(allocator, &[_][]const u8{
        try fs.selfExeDirPath(&buf), default_dictionary,
    });
    defer allocator.free(file_path);
    const file = try fs.cwd().openFile(file_path, .{});
    defer file.close();
    const stat = try file.stat();

    if (stat.size == 0)
        return error.InvalidFormat;

    return os.mmap(null, stat.size, os.PROT.READ, os.MAP.PRIVATE, file.handle, 0);
}

// zig fmt: off
pub fn writePassphrase(
    writer: anytype,
    prng: anytype,
    words: []const u8,
    options: Options
) !void {
// zig fmt: on
    try writeRandomWords(writer, prng, words, options);

    if (options.add_symbol)
        try writeRandomSymbol(writer, prng);

    if (options.add_digit)
        try writeRandomDigit(writer, prng);
}

pub fn writePassword(writer: anytype, prng: anytype, letters: u8) !void {
    var rem = letters;

    while (rem != 0) : (rem -= 1) {
        _ = try writer.writeByte(prng.intRangeAtMost(u8, '!', '~'));
    }
}

// zig fmt: off
pub fn writeRandomWords(
    writer: anytype,
    prng: anytype,
    words: []const u8,
    options: Options
) !void {
// zig fmt: on
    var rem = options.words;

    while (rem != 0) : (rem -= 1) {
        const word = try findWordSlice(words, prng);

        if (options.capitalize) {
            var buf: [4]u8 = undefined;
            const letter_len = try std.unicode.utf8ByteSequenceLength(word[0]);
            const first_letter = try std.unicode.utf8Decode(word[0..letter_len]);
            const upper_len = try std.unicode.utf8Encode(toUpper(first_letter), &buf);
            try writer.writeAll(buf[0..upper_len]);
            try writer.writeAll(word[letter_len..]);
        } else try writer.writeAll(word);

        if (rem != 1)
            if (options.separator) |sep|
                try writer.writeByte(sep);
    }
}

fn findWordSlice(words: []const u8, prng: anytype) ![]const u8 {
    var b: usize = undefined;
    const i = prng.uintLessThan(usize, words.len);
    const e = i + (mem.indexOfScalar(u8, words[i..], '\n') orelse
        return error.InvalidFormat);

    if (@subWithOverflow(usize, e, 1, &b))
        return error.InvalidFormat;

    while (b != 0) : (b -= 1) {
        if (words[b] == '\n') {
            b += 1;
            break;
        }
    }

    return words[b..e];
}

fn writeRandomSymbol(writer: anytype, prng: anytype) !void {
    const symbols = [32]u8{
        '!', '"', '#', '$', '%', '&', '\'', '(',
        ')', '*', '+', ',', '-', '.', '/',  ':',
        ';', '<', '=', '>', '?', '@', '[',  '\\',
        ']', '^', '_', '`', '{', '|', '}',  '~',
    };
    const i = prng.uintLessThan(usize, symbols.len);

    try writer.writeByte(symbols[i]);
}

fn writeRandomDigit(writer: anytype, prng: anytype) !void {
    try writer.writeByte(prng.intRangeAtMost(u8, '0', '9'));
}
