const std = @import("std");
const fs = std.fs;
const io = std.io;
const os = std.os;
const mem = std.mem;
const testing = std.testing;
const allocator = testing.allocator;
const expect = testing.expect;
const expectEqual = testing.expectEqual;
const expectError = testing.expectError;
const pp = @import("main.zig");
const Options = @import("options.zig").Options;
const toUpper = @import("to_upper.zig").toUpper;

const words =
    \\sitjestillingi
    \\pubertetsalder
    \\knivstikkar
    \\profesjonist
    \\matreiddum
    \\randomisera
    \\aksleleddi
    \\lunarisk
    \\hostemikstur
    \\folklorist
    \\familiefest
    \\velkometale
    \\putleteare
    \\samletal
    \\bankgruppe
    \\hypokonder
    \\terskelfjord
    \\illiberalitet
    \\slagbord
    \\utklass
;

const uniwords =
    \\åååå
    \\ääää
    \\öööö
    \\ææææ
    \\þþþþ
;

test "toUpper" {
    const test_pairs = .{
        .{ 'A', 'a' },
        .{ 'A', 'A' },
        .{ 'Å', 'å' },
        .{ '𞤡', '𞥃' },
    };
    inline for (test_pairs) |p|
        try expectEqual(@as(u21, p.@"0"), toUpper(p.@"1"));
}

test "separator" {
    var args = [_][]const u8{ "-s", "_" };
    var opts = Options{};
    try opts.parse(args[0..]);

    if (opts.separator) |sep| {
        try expectEqual(sep, '_');
    } else unreachable;
}

test "invalid separator" {
    var args = [_][]const u8{ "-s", ":-)" };
    var opts = Options{};
    try expectError(error.InvalidValue, opts.parse(args[0..]));
}

test "empty separator" {
    var args = [_][]const u8{ "-s", "" };
    var opts = Options{};
    try opts.parse(args[0..]);
    try expectEqual(opts.separator, null);
}

test "number of words" {
    var out_buf: [0x1000]u8 = undefined;
    var stream = io.fixedBufferStream(out_buf[0..]);
    var xoshiro = std.rand.DefaultPrng.init(0);
    var prng = xoshiro.random();
    var opts = Options{};
    opts.words = 20;
    try pp.writeRandomWords(stream.writer(), &prng, words, opts);
    var wc: u8 = 1;
    for (stream.getWritten()) |b| {
        if (' ' == b)
            wc += 1;
    }
    try expectEqual(opts.words, wc);
}

test "add symbol" {
    var out_buf: [0x1000]u8 = undefined;
    var stream = io.fixedBufferStream(out_buf[0..]);
    var xoshiro = std.rand.DefaultPrng.init(0);
    var prng = xoshiro.random();
    var opts = Options{};
    opts.words = 1;
    opts.add_symbol = true;
    const words2 = "apa\nvad\nlol\n";
    try pp.writePassphrase(stream.writer(), &prng, words2, opts);
    const phrase = stream.getWritten();
    try expectEqual(phrase.len, 4);
    const lb = phrase[3];
    try expect(isSymbol(lb));
}

fn isSymbol(char: u8) bool {
    return ('!' <= char and char <= '/') or (':' <= char and char <= '@') or ('[' <= char and char <= '`') or ('{' <= char and char <= '~');
}

test "add digit" {
    var out_buf: [0x1000]u8 = undefined;
    var stream = io.fixedBufferStream(out_buf[0..]);
    var xoshiro = std.rand.DefaultPrng.init(0);
    var prng = xoshiro.random();
    var opts = Options{};
    opts.words = 1;
    opts.add_digit = true;
    const words2 = "apa\nvad\nlol\n";
    try pp.writePassphrase(stream.writer(), &prng, words2, opts);
    const phrase = stream.getWritten();
    try expectEqual(phrase.len, 4);
    const lb = phrase[3];
    try expect('0' <= lb and lb <= '9');
}

test "capitalize" {
    var out_buf: [0x1000]u8 = undefined;
    var stream = io.fixedBufferStream(out_buf[0..]);
    var xoshiro = std.rand.DefaultPrng.init(0);
    var prng = xoshiro.random();
    var opts = Options{};
    opts.words = 3;
    opts.capitalize = true;
    try pp.writePassphrase(stream.writer(), &prng, words, opts);
    var cc: u8 = 0;
    for (stream.getWritten()) |b| {
        if ('A' <= b and b <= 'Z')
            cc += 1;
    }
    try expectEqual(@as(u8, 3), cc);
}

test "case-fold unicode" {
    var out_buf: [0x1000]u8 = undefined;
    var stream = io.fixedBufferStream(out_buf[0..]);
    var xoshiro = std.rand.DefaultPrng.init(0);
    var prng = xoshiro.random();
    var opts = Options{};
    opts.words = 3;
    opts.capitalize = true;
    opts.separator = null;
    try pp.writePassphrase(stream.writer(), &prng, uniwords, opts);
    var it = (try std.unicode.Utf8View.init(stream.getWritten())).iterator();
    var cc: u8 = 0;
    while (it.nextCodepoint()) |cp| {
        if (toUpper(cp) == cp)
            cc += 1;
    }
    try expectEqual(@as(u8, 3), cc);
}

test "empty wordlist" {
    var out_buf: [0x1000]u8 = undefined;
    var stream = io.fixedBufferStream(out_buf[0..]);
    var xoshiro = std.rand.DefaultPrng.init(0);
    var prng = xoshiro.random();
    var opts = Options{};
    // A completely empty file is rejected when mmapping.
    const words2 = "\n";
    const res = pp.writePassphrase(stream.writer(), &prng, words2, opts);
    try expectError(error.InvalidFormat, res);
}

test "password" {
    var out_buf: [0x1000]u8 = undefined;
    var stream = io.fixedBufferStream(out_buf[0..]);
    var xoshiro = std.rand.DefaultPrng.init(0);
    var prng = xoshiro.random();
    var opts = Options{};
    const len: usize = 24;
    try expect(len != opts.letters);
    try pp.writePassword(stream.writer(), &prng, len);
    try expectEqual(len, stream.getWritten().len);
}
